""" This is our workflow server API. """
import json
import logging
import os
from pathlib import Path

import transaction
import zope.sqlalchemy
from pyramid.config import Configurator
from pyramid.renderers import JSONP
from pyramid.request import Request
from pyramid.response import FileResponse, Response
from pyramid.view import view_config, view_defaults
from pyramid_beaker import session_factory_from_settings

from workspaces.system.schema import get_engine, get_session_factory
from workspaces.workflow.schema import Workflow, WorkflowRequest, WorkflowRequestFile
from workspaces.workflow.services.workflow_info import WorkflowInfo
from workspaces.workflow.services.workflow_service import (
    WorkflowMessageHandler,
    WorkflowService,
)

# pylint: disable=E0401

logger = logging.getLogger(__name__)


# ---------------------------------------------------------
#
#    S E R V I C E S
#
# ---------------------------------------------------------


def lookup_workflow(request: Request) -> Workflow:
    """
    Look up the workflow definition for this workflow name.

    :param request: a workflow request
    :return: the workflow info
    """
    return request.info.lookup_workflow_definition(request.matchdict["name"])


def lookup_request(request: Request) -> WorkflowRequest:
    """
    Look up the workflow info for this request

    :param request: a workflow request
    :return: the workflow info
    """
    return request.info.lookup_workflow_request(request.matchdict["request_id"])


def lookup_file(request: Request) -> WorkflowRequestFile:
    """
    Get the file info for this request

    :param request: a workflow request
    :return: JSON blob containing filename and file content for this request, if any
    """
    return next(
        file
        for file in lookup_request(request).files
        if file.filename == request.matchdict["filename"]
    )


@view_defaults(route_name="workflows", renderer="json")
class WorkflowRestService:
    """
    Top-level service for workflow requests.

    TODO: rename this to requests and add new service about workflow definitions with this name
    """

    def __init__(self, request: Request):
        self.request = request

    @view_config(request_method="GET")
    def list_workflows(self):
        """
        List the workflow requests that we know about.

        Audience: front-end

        :return: None
        """
        return self.request.info.all_workflows()

    @view_config(request_method="GET", route_name="workflow")
    def get_workflow(self):
        """
        Look up a workflow request

        Audience: front-end

        :return: None
        """
        return self.request.context


@view_defaults(route_name="workflow_workdir")
class WorkflowWorkingDirRestService:
    """
    Service for getting working directory files and file content
    """

    def __init__(self, request: Request):
        self.request = request

    @view_config(request_method="GET")
    def show_workdir_content(self):
        """
        List directories and files in a requested workflow's working directory

        Audience: front-end

        :return: None
        """
        requested_workflow = self.request.info.lookup_workflow_request(
            self.request.matchdict["request_id"]
        )
        results_path = requested_workflow.results_dir
        parent_paths = [requested_workflow.results_dir]

        # self.request.matchdict["paths"] is a list of subpaths after dir/ in the route
        # /workflows/{name}/requests/{request_id}/dir/*path
        # Example:
        #       dir/a/b/c == ['a', 'b', 'c']
        for path in self.request.matchdict["paths"]:
            results_path += "/" + path
            parent_paths.append(results_path)

        workdir_dict = self.generate_working_directory_dict(results_path, parent_paths)

        return self.generate_response(workdir_dict)

    @view_config(request_method="GET", route_name="serve_weblog", renderer="json")
    def serve_weblog(self):
        requested_workflow = self.request.info.lookup_workflow_request(
            self.request.matchdict["request_id"]
        )
        results_path = requested_workflow.results_dir
        paths = []
        for root, dirs, _ in os.walk(results_path, followlinks=True):
            for d in dirs:
                if d.startswith("pipeline-"):
                    paths.append(os.path.join(root, d))

        redirect_url = self.generate_url_from_path(paths[0], results_path)
        return Response(status_int=302, location=redirect_url)

    def generate_working_directory_dict(self, results_path, parent_paths) -> dict:
        workdir_dict = {}
        # add 'root' key with requested_workflow.results_dir as the value
        workdir_dict.update({"root": parent_paths[0]})

        # get the last item from the parent_paths list
        # the last item is always the current path
        current_path = parent_paths.pop()
        workdir_dict.update({"current_path": current_path})

        # add 'parents' key with the remaining items in the parent_paths list
        workdir_dict.update({"parents": parent_paths})

        # if the path is a file, return workdir_dict without directory content info
        if Path(results_path).is_file():
            return workdir_dict

        # add 'content' key for housing directory and file information
        workdir_dict.update({"content": {}})
        for key in Path(results_path).iterdir():
            content_key = workdir_dict["content"]

            # check if url needs a slash to divide paths
            divider = ("/", "")[self.request.current_route_url().endswith("/")]
            content_key.update(
                {key.name: {"url": self.request.current_route_url() + divider + key.name}}
            )

            # add full path for content
            content_key[key.name].update({"full_path": key.absolute().__str__()})

            # determine if content is a directory or a file
            content_type = ("directory", "file")[Path.is_file(key)]
            content_key[key.name].update({"content_type": content_type})
        return workdir_dict

    def generate_response(self, workdir_dict):
        # check if current path is a directory
        if Path(workdir_dict.get("current_path")).is_dir():
            # if it is a directory, create a json object
            workdir_json = json.dumps(workdir_dict, indent=2)
            # create response with the json object as the body
            response = Response(body=workdir_json, request=self.request, content_type="text/json")
        else:
            # if it is not a directory, serve the static file
            response = FileResponse(
                workdir_dict.get("current_path"),
                request=self.request,
            )
        return response

    def generate_url_from_path(self, root_path, results_path):
        current_url = self.request.current_route_url()
        return (
            current_url.replace("/weblog", "/dir")
            + root_path.replace(results_path, "")
            + "/html/index.html"
        )


@view_defaults(route_name="workflow_request", renderer="json")
class WorkflowRequestRestService:
    """
    Services for the user-submitted files attached to workflows.
    """

    def __init__(self, request: Request):
        self.request = request

    @view_config(request_method="GET")
    def get_workflow_request(self):
        return self.request.context

    @view_config(request_method="POST", route_name="create_workflow_request")
    def create_workflow_request(self):
        """
        Create a new workflow request from the name/arguments supplied.

        Audience: front-end and CLI
        :return:
        """
        return self.request.info.create_workflow_request(
            self.request.context.workflow_name, self.request.json_body
        )

    @view_config(request_method="POST", route_name="submit_workflow_request")
    def submit_workflow(self):
        """
        Submit this workflow request for processing.

        Audience: front-end and CLI
        :return:
        """
        print(f"Submitting workflow {self.request.context}")

        return self.request.workflows.execute(self.request.context)

    @view_config(request_method="POST", route_name="create_and_submit_workflow_request")
    def create_and_submit(self):
        self.request.context = self.create_workflow_request()
        return self.submit_workflow()

    @view_config(request_method="POST", route_name="announce_qa_ready")
    def announce_qa(self):
        """
        Stopgap for announcing QA ready
        :return:
        """
        print(f"Announcing QA ready for workflow {self.request.context}")

        return self.request.workflows.announce_qa_ready(self.request.matchdict["request_id"])

    @view_config(request_method="POST", route_name="ingest_workflow_result")
    def ingest(self):
        """
        Ingest specified workflow request's results into NGAS and archive
        :return:
        """
        print(f"Ingesting results for workflow request {self.request.context}")

        # 1. retrieve metadata.json for workflow request
        self.request.matchdict["filename"] = "metadata.json"
        file = lookup_file(request=self.request)

        # 2. create ingestion workflow request
        ingest_type = (
            "ingest_cal" if "calibration" in self.request.matchdict["name"] else "ingest_image"
        )
        ingest_request = self.request.info.create_workflow_request(
            workflow=ingest_type,
            argument={"parent_wf_request_id": self.request.matchdict["request_id"]},
        )
        # 3. attach metadata.json to ingestion wf request
        self.request.workflows.attach_file_to_request(
            request=ingest_request, filename=file.filename, content=file.content
        )
        # 4. submit ingestion workflow request
        self.request.workflows.execute(ingest_request)

    @view_config(request_method="POST", route_name="send_carta_url_to_aat")
    def send_carta_url_to_aat(self):
        """
        Pyramid view that sends a given CARTA URL to the AAT's AMQP exchange via the WorkflowService

        return: 200 Response all day, every day
        """
        request_id = self.request.matchdict["request_id"]
        carta_url = self.request.json_body["carta_url"]
        self.request.workflows.send_carta_url_to_aat(request_id, carta_url)

        return Response(status_code=200, body=f"SUCCESS: Sent CARTA URL {carta_url} to AAT system")


@view_defaults(route_name="workflow_request_files", renderer="json")
class WorkflowFilesRestService:
    """
    Services for the user-submitted files attached to workflows.
    """

    def __init__(self, request):
        self.request = request

    @view_config(request_method="PUT", route_name="workflow_request_file")
    def add_file(self):
        """
        Add a file to this workflow request.

        Audience: front-end and CLI
        """
        logger.info(
            "Adding file %s to %s",
            self.request.matchdict["filename"],
            self.request.context,
        )
        file = self.request.workflows.attach_file_to_request(
            request=self.request.context,
            filename=self.request.matchdict["filename"],
            content=self.request.body,
        )
        return file

    @view_config(request_method="GET")
    def get_files(self):
        """
        Get the files associated with this workflow request.

        Audience: front-end
        :return:
        """
        return self.request.context["files"]

    @view_config(
        request_method="GET",
        accept="application/octet-stream",
        route_name="workflow_request_file",
    )
    def get_file_text(self):
        """
        Get the text contents of this file

        Audience: ???
        :return:
        """
        self.request.response.content_type = "application/octet-stream"
        self.request.response.body = lookup_file(self.request).content
        return self.request.response

    @view_config(
        request_method="GET",
        accept="application/json",
        route_name="workflow_request_file",
    )
    def get_file_json(self):
        """
        Get a JSON representation of this file

        Audience: ???
        :return:
        """
        return self.request.context


# ---------------------------------------------------------
#
#    F R A M E W O R K    S E T U P
#
# ---------------------------------------------------------


def get_tm_session(session_factory, transaction_manager):
    """
    Enable Zope's transaction manager on our session
    :param session_factory:
    :param transaction_manager:
    :return:
    """
    dbsession = session_factory()
    zope.sqlalchemy.register(dbsession, transaction_manager=transaction_manager)
    return dbsession


# ---------------------------------------------------------
#
#    M A I N    E N T R Y P O I N T
#
# ---------------------------------------------------------


def main(global_config, **settings):
    with Configurator(settings=settings) as config:
        session_factory = session_factory_from_settings(settings)
        config.set_session_factory(session_factory)
        config.add_renderer("jsonp", JSONP(param_name="callback"))

        settings["tm.manager_hook"] = "pyramid_tm.explicit_manager"

        # use pyramid_tm to hook the transaction lifecycle to the request
        config.include("pyramid_tm")

        # use pyramid_retry to retry a request when transient exceptions occur
        config.include("pyramid_retry")

        session_factory = get_session_factory(get_engine())
        config.registry["dbsession_factory"] = session_factory

        # we need to build a workflow_info here for the message handler, but
        # we won't use it anywhere else, we will make new ones per-request
        workflow_info = WorkflowInfo(get_tm_session(session_factory, transaction.manager))
        message_handler = WorkflowMessageHandler(workflow_info)

        def create_workflow_info(request):
            session = get_tm_session(session_factory, request.tm)
            return WorkflowInfo(session)

        config.add_request_method(create_workflow_info, "info", reify=True)

        # make workflow_service available for use in Pyramid
        config.add_request_method(lambda r: WorkflowService(r.info, message_handler), "workflows", reify=True)

        # GET  /workflows                            <- list of workflows
        # GET  /workflows/null                       <- info about the null workflow
        # POST /workflows/null/requests/create       <- create a request for the null workflow
        # PUT  /workflows/requests/23/files/foo.txt  <- attach foo.txt to request #23 on workflow null
        # POST /workflows/requests/23/submit         <- launch request #23

        config.add_route("workflows", "/workflows")
        config.add_route("workflow", "/workflows/{name}", factory=lookup_workflow)

        # parameters are passed with json: {workflow:<name>, args: <args>}
        config.add_route(
            "create_workflow_request",
            "/workflows/{name}/requests/create",
            factory=lookup_workflow,
        )
        config.add_route(
            "create_and_submit_workflow_request",
            "/workflows/{name}/requests/create-and-submit",
            factory=lookup_workflow,
        )
        config.add_route(
            "workflow_request",
            "/workflows/{name}/requests/{request_id}",
            factory=lookup_request,
        )
        config.add_route(
            "workflow_workdir",
            "/workflows/{name}/requests/{request_id}/dir/*paths",
            factory=lookup_request,
        )
        config.add_route(
            "serve_weblog",
            "/workflows/{name}/requests/{request_id}/weblog",
            factory=lookup_request,
        )
        config.add_route(
            "workflow_request_files",
            "/workflows/{name}/requests/{request_id}/files",
            factory=lookup_request,
        )
        config.add_route(
            "workflow_request_file",
            "/workflows/{name}/requests/{request_id}/files/{filename}",
            factory=lookup_request,
        )

        # parameters are passed with json: {files:{<name>}}.
        # Use empty braces for no additional files
        config.add_route(
            "submit_workflow_request",
            "/workflows/{name}/requests/{request_id}/submit",
            factory=lookup_request,
        )
        config.add_route(
            "ingest_workflow_result",
            "/workflows/{name}/requests/{request_id}/ingest",
            factory=lookup_request,
        )
        # Use this route to send a CARTA envoy URL to the AAT system
        config.add_route(
            "send_carta_url_to_aat",
            "/workflows/carta/requests/{request_id}/send-url-to-aat",
            factory=lookup_request,
        )

        # yes I know this doesn't match pattern, it's a stopgap.
        config.add_route(
            "announce_qa_ready", "/workflows/requests/{request_id}/qa", factory=lookup_request
        )
        config.include("pyramid_beaker")
        config.scan(".")
        return config.make_wsgi_app()
